#!/usr/bin/env python
# -*- coding: utf-8 -*-

# GIMP plugin to create a drawing look using G'mic Diffusion Tensors
#      from a Lyle's flow posted on Gimpchat  

# (c) Dinasset 2020
#
#   History:
#
#   v1.0: 2020-09-30: Initial version
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import sys, os
import time
import datetime
from gimpfu import *
import pygtk
import gtk
from sys import argv, platform
from ctypes import *

# define here some constant value, when required
fixwh=2400
# equates
softlook=0
hardlook=1
fantalook=2
crvname = ["*none","soft chrome","medium chrome","max chrome"]


############################################################################
# include, then use: erro_msg or info_msg
#---------------------------------------------------------------------------
def erro_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(0) # box
    pdb.gimp_message ("ERROR: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###
def info_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2) # error console
    pdb.gimp_message ("INFO: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###
    
############################################################################
# Service routines for GEGL C2G
#---------------------------------------------------------------------------

def load_library (library_name):
    if platform == "linux" or platform == "linux2":
        library_name = library_name + '.so.0'
    elif platform == "win32":
        from ctypes.util import find_library
        library_name = find_library (library_name + "-0")
    else:
        raise BaseException ("TODO")
    return CDLL (library_name)
#import sys
#print (sys.version)
gimp = load_library ('libgimp-2.0')
#print (gimp)
#gimp = load_library ('libgimp-3.0')

def gegl_graph(image, drawable, gegl_graph_string):
    class GeglBuffer(Structure):
        pass
    drawable_id = drawable.ID
    
    gegl = load_library ('libgegl-0.4')
    sucess = gegl.gegl_init (None, None)
    gimp.gimp_drawable_get_shadow_buffer.restype = POINTER (GeglBuffer)
    gimp.gimp_drawable_get_buffer.restype        = POINTER (GeglBuffer)

    x = c_int (); y = c_int (); w = c_int (); h = c_int ()
    non_empty,x,y,w,h = pdb.gimp_drawable_mask_intersect (drawable)
    args = [b"string", c_char_p (gegl_graph_string), c_void_p ()]
    
    if non_empty:
        source = gimp.gimp_drawable_get_buffer (drawable_id)
        target = gimp.gimp_drawable_get_shadow_buffer (drawable_id)
        sucess = gegl.gegl_render_op (source, target, "gegl:gegl", *args)
        gegl.gegl_buffer_flush (target)
        gimp.gimp_drawable_merge_shadow (drawable_id, PushUndo = True)
        gimp.gimp_drawable_update (drawable_id, x, y, w, h)
        gimp.gimp_displays_flush ()

def gegl_c2g(image, drawable):
    radius=300
    samples=4   #4
    iterations=10
    enhance_shadows=True
    gegl_graph_string="gegl:c2g radius=%d samples=%d iterations=%d enhance-shadows=%d" % (radius, samples, iterations, enhance_shadows)
    gegl_graph(image, drawable,gegl_graph_string)
  
# from Original drawable create the Output image
#---------------------------------------------------------------------------
def createOutimage(inImage, inDrawable, widthheight,
                   quota_to_equalize, poster, looktype,
                   quota_to_apply_curves, curves):
     
    owidth = inImage.width
    oheight = inImage.height
    otype = inImage.base_type
    oname = inDrawable.name
    
    if "." in oname:
        split = oname.split(".", 1)
        oname = split[0]
        
    # image name
    Iname = inImage.name
    split = Iname.split("(", 1)
    Iname = split[0]
    end = len(Iname)-2
    Iname = Iname[1:end]    # remove []
    
   # define/calculate the scale factor for the (new) output image -------------------------
    realWidthHeight = owidth+oheight      # real size of the input image
    if widthheight == 0 :   # means the user wants to have output size same as input size
        if realWidthHeight<=maxwh:
            widthheight = realWidthHeight
        else:
            widthheight = maxwh  # max
            gimp.message ("forced max width+height")
    optiWidthHeight = widthheight       # size to work on to get best results
    scale = (optiWidthHeight*1.0)/(realWidthHeight*1.0)      # ratio for scaling

    rwidth=int(owidth*scale)
    rheight=int(oheight*scale)
    
   # create a new image of the decided size ------------------------------------------------
   # --- copy the source layer
   # --- scale
    newImage = pdb.gimp_image_new (rwidth, rheight, otype)
    isamask = pdb.gimp_item_is_layer_mask (inDrawable)
    if isamask==True:    
        inLayer = pdb.gimp_layer_from_mask (inDrawable)
        newBG = pdb.gimp_layer_new_from_drawable (inLayer, newImage)
    else:    
        newBG = pdb.gimp_layer_new_from_drawable (inDrawable, newImage)
    newBG.mode = LAYER_MODE_NORMAL
    newBG.name = "NewImage: Background Layer from "+oname
    newBG.opacity = 100.0 
    newImage.add_layer(newBG, 0)
    pdb.gimp_layer_scale(newBG, rwidth, rheight, False)
    pdb.gimp_layer_add_alpha (newBG)    # RGBA
    newImage.filename=Iname
    (mean,B,C,D,E,F) = pdb.gimp_drawable_histogram (newBG, HISTOGRAM_LUMINANCE, 0.0, 1.0)
    #gimp.message ("mean="+str(mean))
    if mean<96.0:
        pdb.gimp_drawable_levels (newBG, HISTOGRAM_VALUE, 0.0, 1.0, False,
                                      1.10, 0.0, 1.00, False)
        # remove black values 0.0-16.0
        pdb.gimp_drawable_levels (newBG, HISTOGRAM_VALUE, 0.0, 1.0, False,
                                      1.10, 0.0625, 1.00, False)  
    elif mean>120.0:
        pdb.gimp_drawable_levels (newBG, HISTOGRAM_VALUE, 0.0, 1.0, False,
                                      0.90, 0.0, 1.00, False)
        # remove white values 240.0-255.0
        pdb.gimp_drawable_levels (newBG, HISTOGRAM_VALUE, 0.0, 1.0, False,
                                      0.90, 0.0, 0.9375, False)  
    
   # keep a copy of the original background below the one which will be used ----------------
   # --- and set its visibility off (it should not enter in the process)
    Original = newBG.copy()
    newImage.add_layer(Original, 1)     # below
    pdb.gimp_item_set_visible (Original, False)
    Original.name = oname+" (original)"
    
    # check whether there is a mask
    usermask = pdb.gimp_layer_get_mask (newBG)
    #pdb.gimp_message (str(usermask))
    if usermask<>None:
        maskedlayer=True
        OriginalMasked=Original.copy() # this is masked
        newImage.add_layer (OriginalMasked, 2)
        OriginalMasked.name = oname+" (original with mask)"
        pdb.gimp_layer_remove_mask (newBG, MASK_DISCARD)
        pdb.gimp_image_remove_layer (newImage, Original) # was masked
        Original = newBG.copy()                          # redo to get it unmasked
        newImage.add_layer (Original, 1)
        Original.name = oname+" (original not masked)"
    else:
        maskedlayer=False
        OriginalMasked=None
    
   # add to the BG some "equalization" of the colours values ------------
   # --- hence duplicate the BG, apply the gimp equalize, set the opacity to quota_to_equalize,
   # --- merge down with the BG
    if int(quota_to_equalize)>0:  
        newBGequ = newBG.copy()
        newImage.add_layer(newBGequ,0)
        pdb.gimp_drawable_equalize (newBGequ, True)
        newBGequ.opacity=float(quota_to_equalize*1.0)   # float reqd  
        newBG=pdb.gimp_image_merge_down (newImage, newBGequ, CLIP_TO_IMAGE)
        newBG.name = newBG.name+" with partially equalized colours"
    if poster==True:
        pdb.gimp_drawable_posterize (newBG, 9)
        newBG.name = newBG.name+" posterized"
    #if looktype==fantalook: # only for fantasy there is an optional curve to apply
    if curves>0:    #not "*none"
        newBGcrv = newBG.copy()
        newImage.add_layer(newBGcrv,0)
        if curves==1:   #"soft chrome"
            pdb.gimp_drawable_curves_spline (newBGcrv, HISTOGRAM_VALUE, 14,
                                    (0.00,0.00, 0.11,0.62, 0.37,0.25, 0.50,0.86, 0.75,0.64,
                                     0.87,0.99, 1.00,1.00) )
        if curves==2:   #"medium chrome"
            pdb.gimp_drawable_curves_spline (newBGcrv, HISTOGRAM_VALUE, 14,
                                    (0.00,0.00, 0.13,0.75, 0.32,0.12, 0.50,1.00, 0.75,0.37,
                                     0.87,1.00, 1.00, 0.73) ) 
        if curves==3:   #"max chrome"
            pdb.gimp_drawable_curves_spline (newBGcrv, HISTOGRAM_VALUE, 18,
                                    (0.00,1.00, 0.125,0.00, 0.25,1.00, 0.375,0.00, 0.50,1.00,
                                     0.625,0.00, 0.75,1.00, 0.875,0.00, 1.00, 1.00) )
        newBGcrv.opacity=float(quota_to_apply_curves*1.0)   # float reqd  
        newBG=pdb.gimp_image_merge_down (newImage, newBGcrv, CLIP_TO_IMAGE)
        newBG.name = newBG.name+" with curve "+str(crvname[curves])
            
    newBG.opacity = 100.0
    
    return(newImage, maskedlayer, OriginalMasked)

#--------------------------------------------------------------------------- 
    

############################################################################
# Main
#---------------------------------------------------------------------------

def DrawStyleC (inImage, inLayer, outimgWH, poster, curves,
                drawcolour, looktype, texturesize,
                doFlatten):
    
    
    # here the logic of the filter
    oname = inLayer.name
    owidth = inLayer.width
    oheight = inLayer.height
    
    if "." in oname:
        split = oname.split(".", 1)
        oname = split[0]
    
    ts = time.time()
    st = datetime.datetime.fromtimestamp(ts).strftime("%Y-%m-%d %H-%M-%S")
    
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2) # error console
    
    if looktype==softlook: quota_to_equalize = 5; quota_to_apply_curves = 30
    if looktype==hardlook: quota_to_equalize = 10; quota_to_apply_curves = 40
    if looktype==fantalook: quota_to_equalize = 15; quota_to_apply_curves = 50
    
    if looktype<>fantalook: # only for fantasy there is an optional curve to apply
        if curves>0:        #" not *none"
            info_msg ("Curves should be applied mainly for fantasy look ")
    
    (newImage, maskedlayer, OriginalMasked) = createOutimage (inImage, inLayer,
                                              fixwh, quota_to_equalize, poster,
                                              looktype, quota_to_apply_curves, curves)
    Preprocessed = newImage.layers[0]
    Preprocessed.name = "Preprocessed"
    
    nwidth = Preprocessed.width
    nheight = Preprocessed.height

    
    Posterized = Preprocessed.copy()
    newImage.add_layer (Posterized, 0)
    Posterized.name = Preprocessed.name+" (posterized)"
    pdb.gimp_drawable_posterize (Posterized, 15)
    
    #pdb.gimp_display_new (newImage)
    #return
    
    # apply some texture (to avoid flat areas)
    #[G'MIC] Add Grain: fx_simulate_grain 4,1,0.2,100,0,0,0,0,0,0,0,0
    pdb.plug_in_gmic_qt(newImage, Posterized, 1,2,
        "-v - -fx_simulate_grain 3,1,0.2,100,0,0,0,0,0,0,0,0 ")
    Grain = newImage.layers[0]
    Grain.name = "Textured by Grain"
    Grain.mode = LAYER_MODE_HSV_VALUE
    if looktype==softlook: Grain.opacity = 55.0
    if looktype==hardlook: Grain.opacity = 90.0
    if looktype==fantalook: Grain.opacity = 95.0
    
    #[G'MIC] Camouflage: fx_camouflage 2,32,500,90,90,90,0,150,150,150,150,0,255,0,0
    pdb.plug_in_gmic_qt(newImage, Posterized, 1,2,
       #"-v - -fx_camouflage 2,32,100,30,45,60,60,120,90,180,120,60,240,240,160 ")
        "-v - -fx_camouflage 2,32,500,90,90,90,0,150,150,150,150,0,255,0,0 ")
    Camouflage = newImage.layers[0]
    Camouflage.name = "Texture by camouflage"
    Camouflage.mode = LAYER_MODE_GRAIN_MERGE
    if looktype==softlook: Camouflage.opacity = 45.0
    if looktype==hardlook: Camouflage.opacity = 70.0
    if looktype==fantalook: Camouflage.opacity = 95.0
    
    Textured = pdb.gimp_layer_new_from_visible (newImage, newImage, "")
    newImage.add_layer (Textured, 0)
    Textured.name = "Textured by G'MIC Grain+Camouflage"
    Textured.mode = LAYER_MODE_NORMAL
    Textured.opacity = 100.0
    
    #pdb.gimp_display_new (newImage)
    #return

    # scale the layer temporarily
    if texturesize==0: divisor=1.0
    if texturesize==1: divisor=1.414
    if texturesize==2: divisor=2.0
    pdb.gimp_layer_scale (Textured, int(nwidth/divisor), int(nheight/divisor), True)
    
    if looktype==softlook:
        pdb.plug_in_gmic_qt(newImage, Textured, 1,2,
            "-v - -fx_diffusiontensors 10,5,3,1,0.15,1,0,3,0,50,50 ")
    if looktype==hardlook:
        pdb.plug_in_gmic_qt(newImage, Textured, 1,2,
            "-v - -fx_diffusiontensors 10,5,3,2,0.15,1,0,3,0,50,50 ")
    if looktype==fantalook:
        pdb.plug_in_gmic_qt(newImage, Textured, 1,2,
            "-v - -fx_diffusiontensors 10,5,2,2,0.15,1,0,3,0,50,50 ")
    Tensors = newImage.layers[0]
    Tensors.name = "Draw tensors thru G'MIC diffusion tensors"
      
    if looktype==softlook:
        Tensors.mode = LAYER_MODE_HSV_VALUE
        Tensors.opacity = 50.0
    if looktype==hardlook:
        Tensors.mode = LAYER_MODE_HSV_VALUE
        Tensors.opacity = 60.0
    if looktype==fantalook:
        Tensors.mode = LAYER_MODE_GRAIN_MERGE
        Tensors.opacity = 70.0
        
    #pdb.gimp_display_new (newImage)
    #return
        
    # scale back the layer  
    pdb.gimp_layer_scale (Textured, nwidth, nheight, True)
    pdb.gimp_layer_scale (Tensors, nwidth, nheight, True)
    
    pdb.plug_in_spread (newImage, Tensors, nwidth/300.0, nheight/300.0) ####
        
    #pdb.gimp_display_new (newImage)
    #return
    
    FirstOutcome = pdb.gimp_layer_new_from_visible (newImage, newImage, "Visible")
    newImage.add_layer (FirstOutcome, 0)
    FirstOutcome.mode = LAYER_MODE_NORMAL
    FirstOutcome.opacity = 100.0
    FirstOutcome.name = "First Outcome: Tensors merged Smoothed "
    if looktype==softlook:
        pdb.gimp_drawable_curves_spline (FirstOutcome, HISTOGRAM_VALUE,
                                         6, (0.0,0.0, 0.5,0.4, 1.0,1.0))
    
    FirstOutcome2 = FirstOutcome.copy()
    newImage.add_layer (FirstOutcome2, 0)
    FirstOutcome2.name = "First Outcome: Tensors merged Smoothed - B&W"
    gegl_c2g(newImage, FirstOutcome2)
    
    if drawcolour==0:
        FirstOutcome2.opacity =  5.0
    if drawcolour==1:
        FirstOutcome2.opacity = 95.0
    if drawcolour==2:
        FirstOutcome2.opacity = 50.0
    
    #pdb.gimp_display_new (newImage)
    #return
    
    # duplicate the Preprocessed on Top to get better colouring
    PreprocessedOnTop = Preprocessed.copy()
    newImage.add_layer (PreprocessedOnTop, 0)
    if looktype==softlook:
        PreprocessedOnTop.mode = LAYER_MODE_LCH_COLOR   
        PreprocessedOnTop.opacity = 100.0
    if looktype==hardlook:
        PreprocessedOnTop.mode = LAYER_MODE_LCH_COLOR   
        PreprocessedOnTop.opacity = 100.0
    if looktype==fantalook:
        PreprocessedOnTop.mode = LAYER_MODE_SOFTLIGHT   
        PreprocessedOnTop.opacity = 100.0
    PreprocessedOnTop.name = "PreprocessedOnTop"
    
    PreprocessedOnTop2 = Preprocessed.copy()
    newImage.add_layer (PreprocessedOnTop2, 0)
    PreprocessedOnTop2.name = "PreprocessedOnTop 2"
    
    if drawcolour>0: gegl_c2g(newImage, PreprocessedOnTop)
    if drawcolour==1: gegl_c2g(newImage, PreprocessedOnTop2)  # full gray
    PreprocessedOnTop2.mode = LAYER_MODE_SCREEN
    if looktype==softlook: PreprocessedOnTop2.opacity = 20.0
    if looktype==hardlook: PreprocessedOnTop2.opacity = 15.0
    if looktype==fantalook: PreprocessedOnTop2.opacity = 10.0
    if drawcolour==2:
        PreprocessedOnTop2.mode = LAYER_MODE_LCH_COLOR
        PreprocessedOnTop2.opacity = 75.0   # to get more colour
    
    #pdb.gimp_display_new (newImage)
    #return
    
    # final step
    option = ", with a draw effect style C"
    if drawcolour>0:
        option = option + " (GEGL desat C2G)"
        if drawcolour==1: option = option + " (b&w)"
        else: option = option + " (partial colour)"
    else:
        option = option + " (colour)"
    
    visible = pdb.gimp_layer_new_from_visible (newImage, newImage, "visible")
    newImage.add_layer (visible, 0)
        
    # re-apply tensors on top (std)
    FinalTensors = Tensors.copy()
    newImage.add_layer (FinalTensors, 0)
    FinalTensors.name = "Final tensors"
    FinalTensors.mode = LAYER_MODE_OVERLAY
    if looktype==softlook: FinalTensors.opacity = 40.0
    if looktype==hardlook: FinalTensors.opacity = 60.0
    if looktype==fantalook: FinalTensors.opacity = 50.0
    FinalTensors2 = FinalTensors.copy()
    newImage.add_layer (FinalTensors2, 0)
    FinalTensors2.name = "Final tensors B&W"
    FinalTensors2.mode = LAYER_MODE_GRAIN_MERGE
    gegl_c2g(newImage, FinalTensors2)
    if drawcolour==1: gegl_c2g(newImage, FinalTensors)
    
    if drawcolour==0:
        FinalTensors2.opacity = FinalTensors.opacity*0.30
        FinalTensors2.mode = LAYER_MODE_LINEAR_BURN
    if drawcolour==1: # B&W
        #pdb.plug_in_vinvert (newImage, FinalTensors2)
        FinalTensors2.opacity = FinalTensors.opacity*0.40
        FinalTensors2.mode = LAYER_MODE_LINEAR_BURN
    if drawcolour==2:
        FinalTensors2.opacity = FinalTensors.opacity*0.50
        FinalTensors2.mode = LAYER_MODE_LINEAR_BURN
    # check mask
    if maskedlayer==True:
        pdb.gimp_image_raise_item_to_top (newImage, OriginalMasked)
        pdb.gimp_item_set_visible (OriginalMasked, True)
        originalmask = pdb.gimp_layer_get_mask (OriginalMasked)
        pdb.plug_in_vinvert (newImage, originalmask)
    
    if doFlatten==True:
        outcome = pdb.gimp_image_merge_visible_layers (newImage, CLIP_TO_BOTTOM_LAYER)
    else:    
        outcome = pdb.gimp_layer_new_from_visible (newImage, newImage, "Outcome")
        newImage.add_layer (outcome, 0)
        pdb.gimp_item_set_visible (outcome, False)
        
    outcome.name = st+oname+option
    looks=["soft","hard","fantasy"]
    outcome.name = outcome.name+" ("+str(looks[looktype])+" look)"
    
    if curves>0: outcome.name = outcome.name+" (applied curve "+str(crvname[curves]+")")
    
    if texturesize==0: outcome.name = outcome.name+" (small texture)"
    if texturesize==1: outcome.name = outcome.name+" (medium texture)"
    if texturesize==2: outcome.name = outcome.name+" (large texture)"
    
    iratio = outimgWH*1.0/(newImage.width+newImage.height)*1.0
    if outimgWH==0: # original size
        pdb.gimp_image_scale (newImage, inImage.width, inImage.height)
                              # back to original size
    else:
        iratio = float(outimgWH*1.0/(newImage.width+newImage.height)*1.0)
        pdb.gimp_image_scale (newImage, int(newImage.width*iratio), int(newImage.height*iratio))
                              # out to user chosen size
        
    pdb.gimp_display_new (newImage)
    
    return (outcome)

### Registration
whoiam="\n"+os.path.abspath(sys.argv[0])
author="dinasset (inspired by Lyle)"
menu="<Image>/Diego/Drawing/"
copyrightYear="2020"
desc="Draw Effect Style C (with GEGL)"

register(
    "DrawStyleC",
    desc+whoiam,desc,
    author,author,
    copyrightYear,
    "Draw Effect Style C (with GEGL) ...",
    "*",
    [
        (PF_IMAGE, "image", "Input image", None),
        (PF_DRAWABLE, "drawable", "Input drawable", None),
        (PF_ADJUSTMENT, "outwh", "Output image Width+Heigth"+
                        "\n (0=same as input)", 4000, (0, 10000, 100)),
        (PF_TOGGLE, "poster", "Posterize source image?", True),
        (PF_OPTION, "curves", "Curves to apply to source image"+
                        "\n (used mainly with fantasy look)", 0, ("*none",
                                                                  "soft chrome",
                                                                  "medium chrome",
                                                                  "max chrome")),
        (PF_OPTION, "drawcol", "Draw colour", 0, ("colour",
                                                   "B&W",
                                                   "partial colour")),
        (PF_OPTION, "looktype", "Draw look type"+"\n", 0, ("soft",
                                                           "hard",
                                                           "fantasy (for children?)")),
        (PF_OPTION, "tsize", "Texture size"+"\n", 1, ("small",
                                                      "medium",
                                                      "large")),
        (PF_TOGGLE, "flatten", "Flatten Image?", False),
    ],
    [
        (PF_DRAWABLE, "odrawable", "Output drawable", None),
        ],
    DrawStyleC,
    menu=menu
)

main()
